#include "cppm_out.h"

static const uint8_t cppm_out::channelOrder[6] = {1, 2, 0, 3, 4, 5}; // Spektrum channel order (array order is {Ail Ch, Ele Ch, Thr Ch, Rud Ch, Gear Ch, Aux Ch} only Ail, Ele, Thr are actually used) 
// static const uint8_t cppm_out::channelOrder[6] = {0, 1, 2, 3, 4, 5}; // Futaba channel order
// pulse spacings in half-microseconds. Throttle channel is throttle so is preset to low, all values 600 less to account for 300 uS pulse
volatile uint16_t cppm_out::pulseWidth[6] = {1400, 2400, 2400, 2400, 2400, 2400}; // Spektrum order 
// volatile uint16_t cppm_out::pulseWidth[6] = {2400, 2400, 1400, 2400, 2400, 2400}; // Futaba order

void cppm_out::start(void) {
  // start driving six-channel cppm on D3 and D13
  pinMode(3, OUTPUT);
  pinMode(13, OUTPUT);
    
  // initialize counter 2 for pulse output duties
  TCCR2A = 0x02; // CTC mode (Clear Timer on Compare Match)
  TCCR2B = 0x02; // prescale = clk/8: count at 2MHz
  OCR2A = 0xFF; // default count target (256 counts per cycle)
  TIMSK2 = 0x02; // Compare Match A Interrupt Enable
}

// driving 6-channel cppm on Digital 3 (PD3) and Digital 13 (PB5) using Timer 2 interrupts - ceptimus 2015-06-25
// pulses are all 300uS wide. From one +ve edge to next separated by 800uS to 2200uS.  Frame is 20mS long
ISR(TIMER2_COMPA_vect) {
  static uint8_t cycles = 0;
  static uint8_t finalCount;
  static uint8_t stage = 0;
  static uint16_t syncWidth = 35800; // whole frame is 20,000 uS.  subtract 7 x 300 uS for the 7 pulses in a 6-channel frame
  if (!cycles) { // time to switch an output
    uint16_t width;
    if (!(stage & 0x01)) { // at even stages, start 300 uS pulse
      PORTD |= 0x08; // switch 3 high - low with high pulses - SPEKTRUM DM8 module requires this
      PORTB &= 0xDF; // switch 13 low - high with low pulses - same as JR X388S buddy output
      // sei(); // allow other interrupts once the output has been changed - BUG - need to write to OCR2A before enabling interrupts to avoid glitches
      width = 600; // pulse time in half-micro seconds
    }
    else { // at odd stages end 300 uS pulse
      PORTD &= 0xF7; // switch 3 low
      PORTB |= 0x20; // switch 13 high
      if (stage == 13) { // final stage so now output the sync pulse
        // sei(); // allow other interrupts once the output has been changed
        width = syncWidth;
        syncWidth = 35800;  // 50Hz frame is 40,000 counts wide, but subtract the 7 x 300 uS pulses
      }
      else {
        width = cppm_out::pulseWidth[stage / 2];
        // sei(); // allow other interrupts once the output has been changed and pulseWidth read
        syncWidth -= width;
      }
    }
    if ((width & 0x00FF) < 50) { // if finalCount would be less than 50 with the default target at 0xFF
      uint8_t adjust = (50 - (width & 0x00FF)) / (width >> 8) + 1;
      OCR2A = 0x100 - adjust;
      cycles = width >> 8;
      finalCount = (width & 0xFF) + adjust * cycles - 1;
    }
    else { // okay to use default 0xFF count value for all but last cycle
      OCR2A = 0xFF;
      cycles = width >> 8;
      finalCount = (width & 0xFF) - 1;
    }  
    if (++stage > 13) { // move to next stage (14 stages for the 14 pulse edges)
        stage = 0;
    }
  }
  else if (cycles-- == 1) { // last cycle, load the remaining count target
    OCR2A = finalCount;
  }
}

void cppm_out::set(uint8_t n, int16_t p) {
  n = channelOrder[n];
  p += 2400; // nominal ppm (like edge to like edge) width is 1500 uS, but that includes the 300 uS  pulse
  if (p < 1000) { // limit range to 800uS - 2200uS ( x 2 - 600 to convert to counts less 600 for pulse)
    p = 1000;
  }
  else if (p > 3800) {
    p = 3800;
  }
  volatile uint16_t *ptr;
  ptr = pulseWidth + n;
  uint8_t SaveSREG = SREG;
  cli();
  *ptr = p;
  // sei();
  SREG = SaveSREG; // this is more complicated than just cli() and sei() but allows this function to be called from main program OR from within another interrupt
}

void cppm_out::set_uS(uint8_t n, int16_t uS) {
  set(n, (uS - 1500) * 2);
}

int16_t cppm_out::get_uS(uint8_t n) {
  n = channelOrder[n];
  int16_t p = (int16_t)pulseWidth[n] - 2400;
  return p / 2 + 1500;
}


