#include "eeprom.h"

// ceptimus 2015-09-07

globalData eeprom::gd;
modelData eeprom::md;
char eeprom::modelName[14];
uint8_t eeprom::buffer[37];

void eeprom::compressName(void) { // compress the 13-char string md.modelName to packed 7-bit fields in first 12 elements of buffer
  uint8_t left, right;
  for (uint16_t i = 0; i < 13; i++) {
    uint16_t locn = i * 7;
    uint8_t index = locn / 8;
    uint8_t posn = locn % 8;
    left = right = (uint8_t)md.modelName[i];
    if (posn == 0) {
      left <<= 1;
      right = 0;
    } else {
      left >>= posn - 1;
      right <<= 9 - posn;
    }
    buffer[index] |= left;
    buffer[index + 1] |= right;
  }
}

void eeprom::decompressName(char *p) { // extract 13-char string from packed 7-bit fields in first 12 elements of buffer
  for (uint16_t i = 0; i < 13; i++) {
    uint16_t locn = i * 7;
    uint8_t index = locn / 8;
    uint8_t posn = locn % 8;
    uint8_t left = ((uint8_t)(buffer[index] << posn)) >> 1;
    uint8_t right = buffer[index + 1] >> 9 - posn;

    p[i] = '\0' + (left | right);
  }
  p[13] = '\0';
}

void eeprom::cInt(int16_t value, uint8_t bits, uint8_t index, uint8_t posn) {  // compress <bits> bits from value.  Dest (start) buffer[index], Dest (start) bit posn <posn>
  uint8_t destMask = 0x01 << posn;
  for (int16_t sourceMask = 0x01 << (bits - 1); sourceMask; sourceMask >>= 1) {
    if (value & sourceMask) {
      buffer[index] |= destMask;
    } // no need to clear bits as they are initialised to 0 in compress()
    destMask >>= 1;
    if (!destMask) {
      destMask = 0x80;
      index++;
    }
  }
}

int16_t eeprom::dInt(uint8_t bits, uint8_t index, uint8_t posn) {  // reverse of cInt
  int16_t value = 0;
  uint8_t sourceMask = 0x01 << posn;
  for (int16_t destMask = 0x01 << (bits - 1); destMask; destMask >>= 1) {
    if (buffer[index] & sourceMask) {
      value |= destMask;
    }
    sourceMask >>= 1;
    if (!sourceMask) {
      sourceMask = 0x80;
      index++;
    }
  }
  return value;
}

void eeprom::compress(void) { // compress modelData structure md into buffer
  for (uint8_t i = 0; i < 37; i++) {
    buffer[i] = 0;
  }
  compressName();
  cInt(md.rudderCentre[0] - 800, 11, 11, 4);
  cInt(md.rudderCentre[1] - 800, 11, 12, 1);
  cInt(md.rudderRight[0] - 800, 11, 14, 6);
  cInt(md.rudderRight[1] - 800, 11, 15, 3);
  cInt(md.rudderLeft[0] - 800, 11, 16, 0);
  cInt(md.rudderLeft[1] - 800, 11, 18, 5);
  cInt(md.elevatorCentre[0] - 800, 11, 19, 2);
  cInt(md.elevatorCentre[1] - 800, 11, 21, 7);
  cInt(md.elevatorUp[0] - 800, 11, 22, 4);
  cInt(md.elevatorUp[1] - 800, 11, 23, 1);
  cInt(md.elevatorDown[0] - 800, 11, 25, 6);
  cInt(md.elevatorDown[1] - 800, 11, 26, 3);
  cInt(md.motorOff - 800, 11, 27, 0);
  cInt(md.motorOn - 800, 11, 29, 5);
  cInt(md.motorSoftStart, 13, 30, 2);
  cInt(md.buttonSpeed, 10, 32, 5);
  cInt(md.motorRunTime / 10, 10, 33, 3);
  if (md.quickBlipEnabled) {
    buffer[34] |= 0x02;
  } else {
    buffer[34] &= 0xFD;
  }
  buffer[35] = md.Arange;
  buffer[36] = md.Aalarm;  
}

void eeprom::decompress(void) { // decompress buffer into modelData structure md
  decompressName(md.modelName);
  md.rudderCentre[0] = 800 + dInt(11, 11, 4);
  md.rudderCentre[1] = 800 + dInt(11, 12, 1);
  md.rudderRight[0] = 800 + dInt(11, 14, 6);
  md.rudderRight[1] = 800 + dInt(11, 15, 3);
  md.rudderLeft[0] = 800 + dInt(11, 16, 0);
  md.rudderLeft[1] = 800 + dInt(11, 18, 5);
  md.elevatorCentre[0] = 800 + dInt(11, 19, 2);
  md.elevatorCentre[1] = 800 + dInt(11, 21, 7);
  md.elevatorUp[0] = 800 + dInt(11, 22, 4);
  md.elevatorUp[1] = 800 + dInt(11, 23, 1);
  md.elevatorDown[0] = 800 + dInt(11, 25, 6);
  md.elevatorDown[1] = 800 + dInt(11, 26, 3);
  md.motorOff = 800 + dInt(11, 27, 0);
  md.motorOn = 800 + dInt(11, 29, 5);
  md.motorSoftStart = dInt(13, 30, 2);
  md.buttonSpeed = dInt(10, 32, 5);
  md.motorRunTime = 10 * dInt(10, 33, 3);
  md.quickBlipEnabled = buffer[34] & 0x02 ? 1 : 0;
  md.Arange = buffer[35];
  md.Aalarm = buffer[36];
}

void eeprom::limitModelNumber(uint8_t *modelNumber) {
  if (*modelNumber < 1) *modelNumber = 1;
  if (*modelNumber > 27) *modelNumber = 27;
}

void eeprom::getGD(void) {
  EEPROM.get(0, gd);
  limitModelNumber(&(gd.modelNum));
}

void eeprom::putGD(void) {
  limitModelNumber(&(gd.modelNum));
  EEPROM.put(0, gd);
}

void eeprom::getMD(uint8_t modelNumber) {
  limitModelNumber(&modelNumber);
  for (uint8_t i = 0; i < 37; i++) {
    buffer[i] = EEPROM.read(37 * modelNumber - 12 + i);
  }
  decompress();
}

void eeprom::putMD(uint8_t modelNumber) {
  limitModelNumber(&modelNumber);
  compress();
  for (uint8_t i = 0; i < 37; i++) {
    EEPROM.update(37 * modelNumber - 12 + i, buffer[i]);
  }
}

char *eeprom::getModelName(uint8_t modelNumber) {
  limitModelNumber(&modelNumber);
  for (uint8_t i = 0; i < 12; i++) {
    buffer[i] = EEPROM.read(37 * modelNumber - 12 + i);
  }
  decompressName(modelName);
  
  return modelName;
}

