#ifndef eeprom_h
#define eeprom_h

#include <Arduino.h>
#include <EEPROM.h>

// ceptimus 2015-09-07

struct globalData { // test flown version used 15 bytes - allowing 25
  uint8_t modelNum;
  int16_t joyMin[2]; // minimum joystick value x, y
  int16_t joyMax[2];
  int16_t joyCentre[2];
  uint8_t batteryScale; // theoretical value is 160 for a 16V input range (10K 22K divider into 5V input)
  uint8_t lowBatteryAlarm; // in tenths of volts
  uint8_t fiveButton; // non-zero value indicates a digital top/bottom/left/right joystick instead of an analogue one
};

struct modelData {  // test flown version used 37 bytes so there's room for 27 model memories
  char modelName[14]; // 13 chars but pad in case of \0 terminator (13 * 7 bits)
  int16_t rudderCentre[2]; // in microSeconds (nominal 1500 centre) (11 bits) [0] is glide, [1] is power (dual rates)
  int16_t rudderRight[2];
  int16_t rudderLeft[2];
  int16_t elevatorCentre[2];
  int16_t elevatorUp[2];
  int16_t elevatorDown[2];  // not used for push button mode, but travel limit in joystick mode
  int16_t motorOff; // nominally 800 - 1000 uS
  int16_t motorOn;
  int16_t motorSoftStart; // time to reach set speed in milliseconds (13 bits)
  int16_t buttonSpeed; // nominal time for a quarter-turn of virtual escapement in milliseconds (10 bits)
  int16_t motorRunTime; // in seconds (10 bits)
  uint8_t quickBlipEnabled; // non-zero value means enabled (1 bit)
  uint8_t Arange; // for telemetry, in tenths of volts
  uint8_t Aalarm; // for telemetry, in tenths of volts
};

/* compressed form of modelData for EEPROM storage
    |7|6|5|4|3|2|1|0|
00  |modelName[0] | |  // 13 * 7 bits for model name
01  |       [1] |   |
02  |     [2] |     |
03  |   [3] |       |
04  | [4] |     [5] |
05  |   |       [6] |
06  | |         [7] |
07  |         [8] | |
08  |       [9] |   |
09  |    [10] |     |
10  |  [11] |       |
11  |[12] | rCnt[0] |  // 12 * ll bits for rudder and elevator positions (power off and power on positions for centre and both limits of each control)
12  |           |   |
13  |    rCnt[1]    |
14  | |    rRt[0]   |
15  |       |       |
16  |  rRt[1]     | |
17  |    rLft[0]    |
18  |   |   rLft[1] |
19  |         |     |
20  |   eCnt[0]     |
21  |     eCnt[1]   |
22  |     |  eUp[0] |
23  |           |   |
24  |     eUp[1]    |
25  | |    eDn[0]   |
26  |       |       |
27  |    eDn[1]   | |
28  |  Motor off    |  2 * 11 bits for motor on and off
29  |   |  Motor on |
30  |         |     |
31  |Motor softstart|  13 bit motor soft start time
32  |   | button spd|  10 bit button speed
33  |       |       |  10 bit motor run time
34  |motor time |Q|S|  1 bit quick blip enable and one bit spare
35  |  A2 range     |
36  |  A2 alarm     |
 */

class eeprom {
  private:
    static void limitModelNumber(uint8_t *modelNumber);
    static uint8_t buffer[37];
    static void compressName(void);
    static void decompressName(char *);
    static void cInt(int16_t value, uint8_t bits, uint8_t index, uint8_t posn); // compress <bits> bits from value.  Dest (start) buffer[index], Dest (start) bit posn <posn>
    static int16_t dInt(uint8_t bits, uint8_t index, uint8_t posn); // reverse of cInt()
    static void compress(void);
    static void decompress(void);
    
  public:
    static char modelName[]; // temporary copy for getModelName
    static globalData gd;
    static modelData md;
    
    static void getGD(void);
    static void putGD(void);
    static void getMD(uint8_t modelNumber);
    static void putMD(uint8_t modelNumber);
    static char *getModelName(uint8_t modelNumber);
};
#endif

