#include "telemetry.h"
#include <EEPROM.h>
#include "eeprom.h"

char telemetry::errMsg[17];
uint8_t telemetry::rssi;
uint8_t telemetry::a1;
uint8_t telemetry::a2;
boolean telemetry::modelHasTelemetry;
boolean telemetry::present;
boolean telemetry::error;
int16_t telemetry::timeout;

void telemetry::start(void) {
  pinMode(9, INPUT); // serial input (Digital 9 is Port B, bit 1)
  PCMSK0 = 0x02; // set mask to allow (only) digital pin 9 pinchange interrupts on Port B
  PCICR |= 0x01; // allow pinchange interrupts for Port B
  modelHasTelemetry = present = error = false;
  timeout = 1000; // mS
}

// this interrupt handler decodes incoming FrSky serial telemetry data on pin 9
// FrSky module gives out RS232 levels (like +/- 6V) so use 100K inline resistor and rely on ATMega input diode clamps to convert to TTL
// See Atmel application note AVR182 explaining how this is an approved way to do it (they even run mains voltage into a pin via a 1M resistor!)
// Signal is also inverted but this code flips the bits back the right way
// only works at 9600N81 with TCNT0 incrementing every 4 uS (e.g. ATmega328 @ 16 MHz)
// works without delay for character values < 128 so fine for normal ASCII
// when receiving character values > 127, the character isn't processed till the start bit of the following character is received
// so if frame ends with such a character, the frame won't be processed till the beginning of the next frame
// by ceptimus on 2015-09-10
ISR(PCINT0_vect) { // Port B pinchange interrupt
  static uint8_t prev = 0;
  static uint8_t currChar = 0;
  static uint8_t currBits = 0;
  uint8_t now = TCNT0;
  sei(); // allow other interrupts once time has been captured
  uint8_t bits = (((uint8_t)(now - prev)) + 13) / 26; // at 9600 baud with 4 uS TCNT, one bit is 1E6 / (4 * 9600) = 26.04 counts, so half a bit is ~13 counts
  if (!bits || bits > 9) { // detect 0 bits when gap between characters and TCNT0 wrapped exactly, also limit bits to 9 to prevent overflow of (currBits + bits)
    bits = 9;
  }
  prev = now;
  if (!(PINB & 0x02)) { // -ve edge, but inverted RS232 so preceding data bits were low
    if (!currBits) { // start of char so bits includes the start bit (start bit is low)
      currBits = 1; // currBits count does includes the start bit
      bits--; // but don't include start bit in character decode
    }
    currChar >>= bits; // received low bits represent '0' in character.  LSB arrives first, MSB last
    currBits += bits;
  }
  else if (currBits) { // preceding data bits were high
    if (currBits + bits > 9) { // trim stop bit if receiving character values >= 128 (MSB will be high and stop bit is also high)
      bits = 9 - currBits;
    }
    currChar = (0xFF00 | currChar) >> bits; // received high bits represent '1' in character.  LSB arrives first, MSB last
    currBits += bits;
  }
  if (currBits > 8) { // start bit plus character received
    uint8_t rxChar = currChar; // don't know if interrupt is re-entrant - if not then this won't help but if it is then allows next bit(s) to be received while handling char
    currChar = currBits = 0;
    telemetry::handleRxChar(rxChar);
  }
}

void telemetry::handleRxChar(uint8_t b) { // decode FrSky basic telemetry data
  static uint8_t packetPosition = 0;
  static uint8_t packet[9];
  static bool byteStuffing = false;

  if (b == 0x7E) { // framing character
    if (packetPosition == 9) {
      handlePacket(packet);
    }
    packetPosition = 0;
    return;
  }

  if (b == 0x7D) {
    byteStuffing = true;
    return;
  }

  if (byteStuffing) {
    byteStuffing = false;
    if (b != 0x5E && b != 0x5D) {
      packetPosition = 0;
      return;
    }
    b ^= 0x20;
  }

  if (packetPosition > 8) {
    packetPosition = 0;
  }
  else {
    packet[packetPosition++] = b;
  }
}

void telemetry::handlePacket(uint8_t *packet) {
  if (packet[0] == 0xFE) { // not interested in telemetry data other than the most basic kind for this application
    a1 = packet[1]; // A1: VCC (Rx voltage) volts = packet[1] * 5.17647E-2; constant is 3.3 * 4 / 255
    a2 = packet[2]; // A2: (usually scaled LiPo voltage)
    rssi = packet[3]; // main (Rx) link quality 100+ is full signal  40 is no signal
    // packet[4] secondary (Tx) link quality.  Strength of signal received by Tx so not particularly useful.  Numbers are about double those of RSSI.
    modelHasTelemetry = present = true;
    cli(); timeout = 1000; sei();
  }
}

void telemetry::check(uint16_t mS) {
  float volts;
  
  if (timeout <= mS) {
    timeout = 0;
    present = false;
  } else {
    timeout -= mS;
  }
  error = false;
  if (!present) {
    strcpy(errMsg, "Telemetry lost! ");
    error = true;
  } else if (rssi <= 45) {
    strcpy(errMsg, "    RSSI Low    ");
    error = true;
  } else {
    if (eeprom::md.Arange == 0) { // using A1
      volts = 3.3 * 4 * a1 / 255;
      
    } else { // using A2
      volts = eeprom::md.Arange / 10.0;
      volts *= a2 / 255;
    }
    int16_t tenths = (int16_t)(volts * 10 + 0.5);
    if (tenths <= eeprom::md.Aalarm) {
      strcpy(errMsg, " Model battery! ");
      error = true;
    }
  }
}


